/**
Macromedia Flash File graphic lib (triangulation for OpenGL).
Copyright (C) 2005  Christophe Choquet (cchoquet@users.sourceforge.net)

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
**/
#include "ogl_graphics.hh"
#include "triangle2.hh"

#include <GL/glx.h>
#include <GL/glu.h>

OglGraphics::OglGraphics(BaseTexture *texture, LocalFontManager *localFontManager) : Graphics(texture,localFontManager) {
}


/**
 * Returns a list of area interior points.
 * Some control points are also inserted to have a better gradient fill effect. They will force Delaunay triangles to share those segments, so the texture coordinate will be set and then interpolated.
 **/
void OglGraphics::getRegions(list<RegionProperty> *regions) {
	needRefine = false;

	for (int p=0; p< nbSegments; p++) {
		list<float>*secondIntersectpointsY = new list<float>;
		getIntersectionPoints(secondIntersectpointsY,regions,p);
		for (list<float>::iterator it = secondIntersectpointsY->begin(); it != secondIntersectpointsY->end(); it++) {
			float yt = *it;
			Segment *seg= &segments[p];
			double xt;
			if (seg->y[0] == seg->y[1]) {
				xt = seg->x[0];
			} else {
				float t2 = double(yt-seg->y[0]) / double(seg->y[1] - seg->y[0]);
			 	xt = seg->x[0]+t2*(seg->x[1]-seg->x[0]);
			}
			float xi = 1e10, xi2 = -1e10;
			int   is=0;
			int  newColorIndex = -1;
			int color=0;
			int leftSegmentColor = -1;
			int lastSegmentColor = -1;
			for (int i=0; i< nbSegments; i++) {
				Segment *s = &segments[i];
				if (s->y[0] > yt && s->y[1] > yt) continue;
				if (s->y[0] < yt && s->y[1] < yt) continue;
				float dy = s->y[1] - s->y[0];
				float t = (yt-s->y[0])/dy;
				if (t<0 && t>1) continue;
				float dx = s->x[1] - s->x[0];
				float tmpxi = s->x[0] + t*dx;
				if (tmpxi > xt && tmpxi < xi) {
					xi = tmpxi;
					is = i;
				}
				if (tmpxi > xi2 && tmpxi < xt-1) {
					if (s->originalIndex[0] != -1 && s->originalIndex[1] != -1) leftSegmentColor = s->colorsIndex[0];
				}
			}
			if (xi == 1e10) continue;
	
			Segment *lastSeg= &segments[is];
			if (seg->originalIndex[0] == -1 && seg->originalIndex[1] == -1) continue;
			if (seg->originalIndex[1] == -1) {						// Enter new area
				newColorIndex = seg->originalIndex[0];
				color = seg->colorsIndex[0];
			}
			if (seg->originalIndex[0] == -1) {						// Leave area 
				newColorIndex = -1;
				color= seg->originalIndex[1];
			}
			if (seg->originalIndex[0] != -1 && seg->originalIndex[1] != -1) {	// Color transition
				newColorIndex = seg->originalIndex[0];
				color = seg->colorsIndex[0];
			}
			if (lastSeg->originalIndex[0] == -1) {						// Leave area 
				lastSegmentColor= lastSeg->originalIndex[1];
			}
			if (seg->originalIndex[0] == -2) {							// Manually closing segment
				newColorIndex = -2;
			}		

			if (fabs(xi-xt) < 1) continue;
			if (newColorIndex >= 0 ) {
				RegionProperty r;
				r.x = (xt+xi)*0.5;
				r.y = yt;			
				FILLSTYLE *f=NULL ;
				f= fillStyles[color];
				r.colorIndex = color;
				r.transparent = false;
				if (f->getFillStyleType() == 0x12) needRefine = true;
				regions->push_back(r);
			} else {
				RegionProperty r;
				r.x = (xt+xi)/2.0f;
				r.y = yt;	
				r.colorIndex = 0; //defaultColorIndex;
				r.transparent = true;
				regions->push_back(r);
			}
	 	}
		secondIntersectpointsY->clear();
		delete secondIntersectpointsY;
	}
}


/**
 * Returns Y intersection coordinates from the current segment
 * to other segments. This is used to add a control point (an intererior fill region attribute)
 * when the current segment intersect overlapping segments.
 **/
void OglGraphics::getIntersectionPoints(list<float> *resY,list<RegionProperty> *regions, int currentSegmentId) {
	Segment *curSeg= &segments[currentSegmentId];
	list<float> tmpY;

	for (int p=0; p< nbSegments; p++) {
		if (p == currentSegmentId) {
			tmpY.push_back(curSeg->y[0]);
			tmpY.push_back(curSeg->y[1]);
		};
		Segment *s = &segments[p];
		if (s->y[0] > curSeg->y[0] && s->y[1] > curSeg->y[1]) continue;
		if (s->y[0] < curSeg->y[0] && s->y[1] < curSeg->y[1]) continue;
		if (s->x[0] > curSeg->x[0] && s->x[1] > curSeg->x[1]) continue;
		if (s->x[0] < curSeg->x[0] && s->x[1] < curSeg->x[1]) continue;
		if (s->x[0] == curSeg->x[0] && s->x[1] == curSeg->x[1]) continue;
		if (s->y[0] == curSeg->y[0] && s->y[1] == curSeg->y[1]) continue;
		if (s->y[1] == s->y[0]) {
			tmpY.push_back(s->y[0]);
		} else {
			float px,py;
			if (segmentIntersect(curSeg->x[0],curSeg->y[0],curSeg->x[1],curSeg->y[1],s->x[0],s->y[0],s->x[1],s->y[1],px,py)) 
				tmpY.push_back(py);
		}
	}
	float y;
	if (curSeg->y[0] < curSeg->y[1]) {
		y = curSeg->y[0];
	} else {
		y = curSeg->y[1];
	}
	for (list<float>::size_type i=0; i< tmpY.size(); i++) { 
		float yc = 1e10;
		for (list<float>::iterator it = tmpY.begin(); it != tmpY.end(); it++) {
			if (*it > y && *it < yc) { yc = *it; } 
		}
		if (yc > y+2) {
			resY->push_back((y+yc)*0.5);
		}
		y = yc;
	}
}

/**
 * Transform the segment list to a set of triangles - OpenGL specific
 **/
DrawingSequence *OglGraphics::shape2Triangles() {
	list<RegionProperty> *regions = new list<RegionProperty>;
	regions->clear();
	getRegions(regions);
	if (regions->size() == 0) { 
		delete regions; 
		return NULL; 
	}

	struct triangulateio in,out,vorout;
    in.pointlist = (double*) malloc(nbPoints*2*sizeof(double));
	for (int i=0; i< nbPoints; i++) {
		in.pointlist[i*2] = points[i*2];
		in.pointlist[i*2+1] = points[i*2+1];
	}
	in.numberofpoints = nbPoints;
	in.numberofpointattributes = 0;
	in.pointattributelist = NULL;
	in.segmentlist = (int*) malloc(nbSegments*2*sizeof(int));
	in.numberofsegments = nbSegments;
	in.segmentmarkerlist = NULL;
	in.numberofregions = 0;
	in.regionlist = NULL;
	in.numberoftriangles = 0;
	for (int i=0; i< nbSegments;i++) {
		in.segmentlist[i*2+0] = segments[i].points[0];
		in.segmentlist[i*2+1] = segments[i].points[1];
	}

	in.numberofholes = 0;
	in.holelist = 0;
	in.pointmarkerlist = 0;
	in.triangleattributelist = 0;

	// Add regions
	in.numberofregions = regions->size();
	in.regionlist = (double*)malloc(4*in.numberofregions*sizeof(double));
	int j=0;	
	list<RegionProperty>::iterator it;
	for (it = regions->begin(); it != regions->end(); it++)  {
		RegionProperty *p = &*it;
		in.regionlist[j*4] = p->x;
		in.regionlist[j*4+1] = p->y;
		in.regionlist[j*4+2] = p->transparent ? -1 : p->colorIndex;
		in.regionlist[j*4+3] = 0; // No area limit specified
		j++;
	}

	out.pointlist = NULL;
	out.trianglelist = NULL;
	out.segmentlist = NULL;
	out.edgelist = NULL;
	out.edgemarkerlist= NULL;
	out.pointmarkerlist = NULL;
	out.segmentmarkerlist= NULL;
	out.pointattributelist = NULL;
	out.regionlist = NULL;
	out.triangleattributelist = NULL;
	out.numberofpoints=0;
	out.numberoftriangles=0;

#if 0
	if (true) {
	    // DUMP FOR TEST
		char tmp[100];
		sprintf(tmp,"/tmp/Triangle/test.poly");
		FILE *f = fopen(tmp,"w");
		fprintf(f,"%d 2 0 1\n",nbPoints);
		for (int i=0; i< nbPoints;i++) fprintf(f,"%d %f %f -1\n",i+1,in.pointlist[i*2],-in.pointlist[i*2+1]);
		fprintf(f,"%d 0\n",nbSegments);
		for (int i=0; i< nbSegments;i++) fprintf(f,"%d %d %d \n",i+1,segments[i].points[0]+1,segments[i].points[1]+1);
		fprintf(f,"0\n");
		fprintf(f,"%d\n",regions->size());
		list<RegionProperty>::iterator it;
		int i = 1;
		for (it = regions->begin(); it != regions->end(); it++)  {
			RegionProperty *p = &*it;
	        fprintf(f,"%d %f %f %d\n",i,p->x,-p->y,p->transparent ? -1 : p->colorIndex);
			i++;
		}
		fclose(f);
	}
#endif
	Triangle triangle;
	char * triswitches;
	if (needRefine) triswitches="-z -p -q10 -Q -A -a5000 -S300"; 
	else triswitches="-z -p -q10 -Q -S500 -A";
	
//	if (needRefine) triswitches="-z -p -Q -A "; 
//	else triswitches="-z -p -Q -A"; 

	triangle.triangulate(triswitches,&in,&out,&vorout);

	//printf("%d %s Register %d points  %d triangles\n",nbFillStyles,triswitches,out.numberofpoints,out.numberoftriangles);
	DrawingSequence	*drawingSequence = new DrawingSequence();
	drawingSequence->allocateVertices(2*out.numberofpoints);
	GLfloat *fpoints = drawingSequence->getVertices();
	for (int i=0; i< out.numberofpoints*2; i++) fpoints[i] = float(out.pointlist[i]);
	free(in.pointlist);
 	drawingSequence->setNbPoints(out.numberofpoints);

	// Optimize indexes to minimize texture changes.
	// Calculate texture coordinates
	for (int i = 0; i< nbFillStyles; i++) {
		int nbTris = 0;
		for (int j=0; j< out.numberoftriangles; j++) if (out.triangleattributelist[j] == i) nbTris++;
		if (nbTris == 0) continue;

		DrawingEntry *drawingEntry = new DrawingEntry();
		drawingEntry->drawingType = DrawingTriangles;
		FILLSTYLE *f = fillStyles[i];
		int fillType = f->getFillStyleType();
		drawingEntry->allocateIndexes(3*nbTris);
		GLuint *indexes = drawingEntry->getIndexes();
		drawingEntry->allocateTexCoords(2*out.numberofpoints);
		GLfloat *texCoords = drawingEntry->getTexCoords();
		int idx = 0;
		int texNum =0;
		float scaleX, scaleY;
		int texDimX, texDimY;
		NORMMATRIX nm;
		switch (fillType) {
			case 0x00: texNum = texture->getColorTexture();
				break;
			case 0x10:
			case 0x12:
				texNum= texture->getGradient(characterId,i);
				nm = normalize(f->getGradientMatrix());
				nm = MatrixInvert(&nm);
				break;
			case 0x40:  // bitmap
			case 0x41:  // clipped bitmap fill
			case 0x42:  // non smoothed repeting bitmap
			case 0x43:  // non smoothed clipped bitmap
				texNum = texture->getExternalTexture(f->getBitmapId(),scaleX,scaleY,texDimX,texDimY);
				nm = normalize(f->getBitmapMatrix());
				nm = MatrixInvert(&nm);
				break;				
		}
		drawingEntry->fillType = fillType;
		drawingEntry->texNum = texNum;
		for (int j=0; j< out.numberoftriangles; j++) {
			if (out.triangleattributelist[j] == i) {
				indexes[idx*3+0] = out.trianglelist[j*3+0];
				indexes[idx*3+1] = out.trianglelist[j*3+1];
				indexes[idx*3+2] = out.trianglelist[j*3+2];
				switch (fillType) { 
					case 0x00: {
						int loc = texture->getColorTextureIndex(f->getColor());
						float tx = (loc%256)/256.0;
						float ty = float(loc/256)/256.0f;
						texCoords[indexes[idx*3+0]*2] = tx; texCoords[indexes[idx*3+0]*2+1] = ty; 
						texCoords[indexes[idx*3+1]*2] = tx; texCoords[indexes[idx*3+1]*2+1] = ty; 
						texCoords[indexes[idx*3+2]*2] = tx; texCoords[indexes[idx*3+2]*2+1] = ty; 
						break;
					}
					case 0x10: { // linear gradient fill
						for (int k=0; k<3; k++) {
							float xl = fpoints[indexes[idx*3+k]*2];
							float yl = fpoints[indexes[idx*3+k]*2+1];
							float a = nm.a *xl + nm.b * yl + nm.translateX + 16384;
							a = a /32768.01;
							float opx = 1.0/512.0;
							if (a<opx) a=opx;
							if (a>1-opx) a=1-opx;
							texCoords[indexes[idx*3+k]] = a;
						}
						break;
					}
					case 0x12: { // radial gradient fill
						for (int k=0; k<3; k++) {
							float xl = fpoints[indexes[idx*3+k]*2];
							float yl = fpoints[indexes[idx*3+k]*2+1];
							float a = (nm.a *xl + nm.b * yl + nm.translateX)/32768;
							float b = (nm.c *xl + nm.d * yl + nm.translateY)/32768;
							float radius = sqrt(a*a+b*b)*2;
							float opx = 1.0/512.0;
							if (radius > 1-opx) radius = 1-opx;
							if (radius < opx) radius = opx;
							texCoords[indexes[idx*3+k]] = radius;
						}
						break;
					}
					case 0x40:  // bitmap
					case 0x41:  // clipped bitmap fill
					case 0x42:  // non smoothed repeting bitmap
					case 0x43: { // non smoothed clipped bitmap
						for (int k=0; k< 3; k++) {
							float xl = fpoints[indexes[idx*3+k]*2];
							float yl = fpoints[indexes[idx*3+k]*2+1];
							float a = nm.a *xl + nm.b * yl + nm.translateX;
							float b = nm.c *xl + nm.d * yl + nm.translateY;
							a = a / scaleX *float(texDimX);
							b = b / scaleY *float(texDimY);
							int ix = int(a);
							if (ix >= texDimX) {
								if (f->getFillStyleType() == 0x41 || f->getFillStyleType() == 0x43) ix = texDimX -1;
								else ix = ix % texDimX;
							}
							int iy = int(b);
							if (iy >= texDimY) {
								if (f->getFillStyleType() == 0x41 || f->getFillStyleType() == 0x43) iy = texDimY-1;
								else iy = iy % texDimY;
							}
							if (ix < 0) {
								if (f->getFillStyleType() == 0x41 || f->getFillStyleType() == 0x43) ix = 0;
								else ix = (-ix)%texDimX;
							}
							if (iy < 0) {
								if (f->getFillStyleType() == 0x41 || f->getFillStyleType() == 0x43) iy = 0;
								else iy = (-iy)%texDimY;
							}
							texCoords[indexes[idx*3+k]*2] = float(ix)/float(texDimX);
							texCoords[indexes[idx*3+k]*2+1] = float(iy)/float(texDimY); 
                        }                       
						break;
					}
				}
				idx++;
			}
		}
		drawingEntry->nbElements = nbTris;
		drawingSequence->addDrawingEntry(drawingEntry);
	}
	free(out.trianglelist);
	free(out.triangleattributelist);
	free(out.regionlist);
	free(out.segmentlist);
	free(out.segmentmarkerlist);
	free(out.pointmarkerlist);
	free(out.pointlist);
	free(out.pointattributelist);
	free(in.segmentlist);
	regions->clear();
	delete regions;
	return drawingSequence;
}
